
#include <stdlib.h>
#include <math.h>
#include "may.h"

#ifdef TRACE
#define PRtr(x) printf x;
#else
#define PRtr(x)
#endif

void circle_map ();

/*-------------------------------------------------------------------*/
/* This routine burns some cpu cycles 
 *
 * This routine computes a scan-line's worth of pixel data.
 * Assembled into a pixmap, the data presents an interesting Moire pattern.
 */

void filled_circle (flt_pt_data, flt_pt_datalen, xloc, yloc)
float  		flt_pt_data [];
unsigned int	flt_pt_datalen;
float		xloc, yloc;
{
   unsigned int	i;
   int		ip, jp;
   
   jp = (int) yloc;
   for (i=0; i<flt_pt_datalen; i++) {
      ip = (int) xloc + (int) i;
      flt_pt_data [i] = ((float) ((ip*ip+jp*jp) % 2500)) / 2500.0;
   }
}

/*-------------------------------------------------------------------*/
/* This routine burns some cpu cycles 
 *
 * This routine computes a scan-line's worth of pixel data.
 * This algorithm generates the very traditional view of the exterior
 * of the Mandelbrot attractor.
 */

void brot_out (flt_pt_data, flt_pt_datalen, 
		re_start, im_start, delta, itermax, roll_over)
float  		flt_pt_data [];
unsigned int	flt_pt_datalen;
double		re_start, im_start, delta;
int		itermax;
int		roll_over;
{
   unsigned int	i,j;
   int		ip, jp;
   double		re_position, im_position;
   double		re, im, tmp;
   int		loop;
   
   im_position = im_start;
   re_position = re_start;
   for (i=0; i<flt_pt_datalen; i++) {
      re = re_position;
      im = im_position;
      for (loop=1; loop <itermax; loop++) {
         tmp = re*re - im*im + re_position;
         im = 2.0*re*im + im_position;
         re = tmp;
         if ((re*re + im*im) > 4.0) break;
      }    
      flt_pt_data [i] = ((float) (loop%roll_over)) / ((float) roll_over); 
      re_position += delta;
   }
}
   
/*-------------------------------------------------------------------*/
/* main
 * main of slave executable.
 *
 * This is the main "slave" process employed in a processor farm.
 * It is implemented as an "event loop", servicing work requests sent to
 * it by the master process.  
 *
 * The basic algorithm embodied here is trivial: read the next work
 * request, perform computation, send reply, loop back.  The actual code
 * is  bit messy because the work requests and replies must be cast into 
 * C structures.
 *
 * In this particular example, scan-lines worth of pixels are computed.
 * Each scan line has length PIX_LINE_LEN
 */

#define PIX_WID 256

void main (argc, argv) 
int argc;
char **argv; 
{

    struct	mayProcId	*myself, *myparent;
    struct 	mayMsghead	header;
    int 	messlen;
    int         retval;		/* returned value */

    struct yall {
        struct initial_data {
	    unsigned int ix;
	    unsigned int iy;
	    unsigned int ilen;
            double fx;
	    double fy;
	    double delta;
	    unsigned int itermax;
	    unsigned int roll_over;
	    } received_data;
        float data [PIX_WID];
        } ball;

#ifdef UDP
    myself = mayInitUdp ();
    myparent = mayGetParentUdp ();
#else
    mayInitTcp ();
    myparent = mayGetParentTcp ();
#endif

    PRtr (("slave: done initializing, about to enter main loop \n"));

    /* set up receiving area */
    header.msgptr = (char *) &ball.received_data;

    /* enter main loop */
    while (TRUE) {

       /* wait up to 20 seconds to receive a message */
       retval =  mayReceive (NULL, NULL, NULL, 20, &header, &messlen);

       /* if we timed out waiting for a message, quit */
       if (retval == MAY_TIMED_OUT) break;

       /* if there was any receive error, terminate */
       if (retval < 0) break;

       /* if we get message 999, quit */
       if (header.msgtype == 999) break;

       if (messlen != sizeof (struct initial_data)) {
          fprintf (stderr, "slave: wrong size message recv'd, ");
          fprintf (stderr, "got length=%d shoulda got length=%d\n",
			messlen, sizeof (struct initial_data) );
       }

       PRtr ((" got the %d th message for %d %d",  header.subtype,
          ball.received_data.ix, ball.received_data.iy ));
       PRtr ((" 	%d %f %f %f %d \n", 
          ball.received_data.ilen,
          ball.received_data.fx,
          ball.received_data.fy,
          ball.received_data.delta,
          ball.received_data.itermax,
          ball.received_data.roll_over));
       PRtr (("Q has %d unread messages in it \n", retval));

       /* burn some CPU cycles computing a picture */
/*
       circle_map (ball.data, ball.received_data.ilen, 
		ball.received_data.fx, ball.received_data.fy,
		ball.received_data.delta, (int) ball.received_data.itermax,
		(int) ball.received_data.roll_over);
*/

       brot_out (ball.data, ball.received_data.ilen, 
		ball.received_data.fx, ball.received_data.fy,
		ball.received_data.delta, (int) ball.received_data.itermax,
		(int) ball.received_data.roll_over);


     /*   (void) filled_circle (ball.data, ball.received_data.ilen, 
		(float) ball.received_data.ix, 
		(float) ball.received_data.iy);  */

       /* send result back to parent process */
       maySend (myparent, (long) header.msgtype, (short) header.subtype,
		(char *) &ball, 
		(int) sizeof (struct yall) ); 

   }

   /* wrap it up & exit */
   PRtr (("Goodbye, World! \n"));
   mayTerminate ();
   exit (EXIT_SUCCESS);
}


/*-------------------------------------------------------------------*/
/* This routine burns some cpu cycles 
 *
 * This routine computes a scan-line's worth of pixel data.
 * This algorithm generates the "measure" or "density" of the visted
 * points in the interior of the Mandelbrot attractor.
 */

extern double drand48 ();
#define LOOP_COUNT 18

void mandelbrot_in (glob, sizex, sizey,
               re_center, im_center, width, itermax)
float  		glob [];
unsigned int	sizex, sizey;
double		re_center, im_center, width;
int		itermax;
{
    unsigned int	i,j, globlen;
   int		ip, jp;
   double		re_start, im_start, delta;
   double		re_position, im_position;
   double		re, im, tmp;
   int		loop;
   double		x_width, y_width;
   double		x_slope, y_slope, x_off, y_off;
   int		horiz_pix, vert_pix;
   
   
   x_width = width;
   y_width = width * ((double) sizey) / ((double) sizex);
   re_start = re_center - x_width / 2.0;
   im_start = im_center - y_width / 2.0;
   x_slope = ((double) sizex) / x_width;
   y_slope = ((double) sizey) / y_width;
   x_off = x_slope * re_start;
   y_off = y_slope * im_start;
   
   
   globlen = sizex*sizey;
   for (i=0; i<globlen; i++) glob [i] = 0.0;
   
   /* very crudely, we want as many random seeds as there are pixels */
   for (i=0; i<globlen*itermax; i++) {
      im_position = y_width * (drand48() - 0.5) + im_center;
      drand48 (); /* throw one away -- absolutely must do this-- dont ask */
      re_position = x_width * (drand48() - 0.5) + re_center;
   
      if (i%(itermax*sizex)==0) printf(" start row %d\n", i/(sizex*itermax));
   
      re = im = 0.0;
      for (loop=0; loop < LOOP_COUNT; loop++) {
         tmp = re*re - im*im + re_position;
         im = 2.0*re*im + im_position;
         re = tmp;
         if ((re*re + im*im) > 4.0) break;
         horiz_pix = (int) (x_slope * re - x_off);
         vert_pix = (int) (y_slope * im - y_off);
         if ( (horiz_pix >= 0) && (horiz_pix < sizex) && 
            (vert_pix >= 0) && (vert_pix < sizey)) {
            glob [vert_pix*sizex + horiz_pix] ++;
         }
      }    
   }
   
   /* renormalize */
   for (i=0; i<globlen; i++) {
      glob [i] = glob[i] / (double) (itermax*LOOP_COUNT) ;
      glob [i] = ((double) (((int) glob[i]) %3)) / 3.0;
   }
}

/*-------------------------------------------------------------------*/
/* This routine burns some cpu cycles 
 *
 * This routine computes a scan-line's worth of pixel data.
 * This algorithm some crazy variant of the circle-map model for a
 * phase-locked loop.
 */

void circle_map (flt_pt_data, flt_pt_datalen, 
		re_start, im_start, delta, itermax, roll_over)
float  		flt_pt_data [];
unsigned int	flt_pt_datalen;
double		re_start, im_start, delta;
int		itermax;
int		roll_over;
{
   unsigned int	i,j;
   int		ip, jp;
   double		re, im, tmp;
   int		loop;
   double re_omega, im_omega;    /* driving frequency */
   double re_K, im_K;            /* coupling strength */
   double re_tmp, im_tmp;

   re_omega = im_omega = 0.0;
   
   im_K = im_start;
   re_K = re_start;
   for (i=0; i<flt_pt_datalen; i++) {
      re = 0.0;
      im = 0.0;
      for (loop=1; loop <itermax; loop++) {

         re_tmp = sin(re) * cosh(im);
         im_tmp = cos(re) * sinh (im);

         re += re_omega + re_K*re_tmp - im_K*im_tmp;
         im += im_omega + im_K*re_tmp + re_K*im_tmp;

         if ((re*re + im*im) > 10.0) break;
      }    
      flt_pt_data [i] = ((float) (loop%roll_over)) / ((float) roll_over); 
      re_K += delta;
   }
}
   
/*-------------------------------------------------------------------*/
