/* =================================================================
 *
 * Copyright (C) 1989, 1990, 1993 Linas Vepstas
 *
 *                   All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its 
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies 
 * and that both the copyright notice and this permission notice 
 * appear in supporting documentation.
 *
 * Permission to redistribute this software and its documentation
 * is hereby granted, provided that such redistribution is not 
 * performed for monetary profit or gain.
 *
 * Redistribution for profit without express written consent is
 * prohibited.  Contact the author if you wish to use this software 
 * and/or its documentation in a commercial product.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND.  
 * THE AUTHOR DISCLAIMS ALL EXPRESS AND IMPLIED WARRANTIES, INCLUDING 
 * BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE HELD LIABLE FOR ANY SPECIAL, DIRECT, 
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING 
 * IN LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Bug reports or fixes may be sent to the author, who may or may not 
 * act upon them as he desires.
 * 
 * The author's last known address is:
 * 
 *                       Linas Vepstas
 *                     1518 Enfield Road
 *                  Austin, Texas 78703-3424
 *                      1-(512)-499-8246
 *                    linas@austin.ibm.com
 *
 * =================================================================
 */

#include "may.h"
#include "maymsg.h"

extern struct hostent *gethostbyname();
struct mayProcId *mayGetHosts ();

/* maintain a list of all open communications channels */
struct mayProcId *open_socks = NULL;	
struct mayProcId *may_hosts = NULL;   /* list of hosts */


/* ============================================================= */
/*
 * mayCreate
 * Executes a remote program
 *
 * This routine tells the maydaemon on the indicated host to start 
 * the execution of the indicated program on that host.  It does so by
 * constructing a "create" message containing the name of the program 
 * and sending it to the maydaemon of the (remote) machine.
 * 
 * Upon creation, this routine also "automatically" opens a connection
 * of the requested type (TCP, UDP, other) to the child process.  The
 * opening of a communication channel is done most efficeintly at creation
 * time, and seems to involve the least amount of elapsed network delay if
 * done at this time.  
 *   
 * At this time, only a TCP create is supported.  If other connections are
 * desired, e.g. UDP, the "mayInitUdp()" routine can be called, and then
 * the UDP addresses can be exchanged via the already open TCP connection.
 * Of course, if all you really wanted was UDP in the first place, this
 * would not be a very efficeint mechanism.  That is why the
 * "requested_connection_type" field is included.  Although this
 * implementation does not support UDP or other, it *could* in principle.
 *   
 * If one wanted to open more than one connection at create time, one can
 * do so by specifying an appropriate requested_connection_type, and
 * having this function return a linked list of ID's !!
 *   
 * De facto, what happens internally is that a TCP socket is openened and
 * connected to the child (via a fork/exec by the daemon).  The TCP socket
 * had to be opened to the daemon to get the create message to the daemon.
 * The daemon connection could have been implemented in another way;
 * choosing TCP seemed like the right thing to do.
 * 
 * psuedocode:
 * >) go out and get the address of the remote daemon.  
 * >) opens a socket.
 * >) tries to connect to daemon socket.
 * >) Constructs the create message, which includes the name of the program
 *    to be executed.  
 * >) sends the create message to the remote daemon.
 * >) waits for a create successful reply.
 * >) returns the socket id to the parent.
 * 
 * Notes:
 * >) uhh, waiting for a create reply is a nice service, but not absolutely
 * necessary.  Maybe we can work something out for greater efficiency, like
 * simultaneous create requests that don't hang waiting for reply ...
 */


#ifdef ANSI_C
struct mayProcId * mayCreate (char *requested_host_name, 
                              char *requested_program,
                              int requested_connection_type)
#else
struct mayProcId * mayCreate (requested_host_name, 
                              requested_program,
                              requested_connection_type)
char *requested_host_name;	/* hostname on which to start */
				/* execution of requested program */
char *requested_program;	/* name of program to be started */
int  requested_connection_type; /* UDP, TCP are supported */
#endif
{

   struct 	mayProcId 	*remhost; /* pointer to structure containing */
					/* info about the requested host */

   struct 	create_msg 	cm;	/* create message */
   struct	create_reply	cr;	/* create reply */
   struct	mayMsghead	header;	/* standard message header */
   int 		len, flag;
   struct 	mayProcId 	*sa;	/* pointer that will be returned */
   struct 	mayProcId 	*tmp;	/* temp use pointer */


   /* go get info about requested host  */
   remhost = mayGetHosts (requested_host_name, "maydaemon");
   if (remhost == NULL) return (NULL);

   /* create socket to connect to daemon */
   remhost->proto = SOCK_STREAM;
   remhost->socket_id = socket (AF_INET, SOCK_STREAM, 0);
   if (0 > remhost->socket_id) {
      perror ("mayCreate: error opening socket to daemon");
      return (NULL);
   }

#ifdef DEBUG
   printf ( " mayCreate connecting to %d %x %d \n", 
       remhost->address.sin_family,
       remhost->address.sin_addr.s_addr, 
       remhost->address.sin_port);
   printf (" mayCreate talking on stream sock fd = %d \n", remhost->socket_id);
#endif

   if (connect (remhost->socket_id, &remhost->address, SIZEOSOX) < 0) {
      fprintf (stderr, 
              "mayCreate: Error connecting to daemon stream socket on host %s: \n", 
               requested_host_name);
      perror ("       ");
      return (NULL);
   }

   /* construct the create message */
   strcpy (cm.prog_name, requested_program);

   if (requested_connection_type == MAY_CREATE_UDP) {
      /* 
       * if this is a UDP create, we need to send the parent's UDP
       * address to the child (via daemon).
       */
      memcpy (&cm.sproc,  The_udp_sock, SIZEOPRO);
   }


   /* fire off the create message to the requested daemon */
   if (0 > maySend (remhost, DAEMONMSG, requested_connection_type,
      (Void *) &cm, sizeof (struct create_msg))) {
      perror ("mayCreate: error sending create message");
      return (NULL);
   }

   /* if we got to here, something must have gone right.
    * Its fairly safe to assume that we will have an active child
    * so lets string up the address.
    */
   if ((tmp = (struct mayProcId *) malloc (SIZEOPRO)) == NULL) {
      perror ("mayCreate: error on malloc");
      return (NULL);
   }
   memcpy (tmp, remhost, SIZEOPRO);
   tmp->unread_bytes_remaining = 0;
   tmp->partial_msg = NULL;

   /* chain this into place so that we can use mayreceive on this
    * address.  For a UDP create, we will have to unchain this after the
    * receive; for a TCP create, this will become permanent */
   tmp->next = open_socks;
   open_socks = tmp;

#ifdef DEBUG
   printf(" maycreate hanging out, waiting for reply\n");
#endif

   /* wait for the create reply message */
   len = sizeof (struct create_reply);
   header.msgptr = (Void *) &cr;
   may_internal_flag = MAY_INTERNAL_FLAG;
   flag = mayReceive (NULL, DAEMONMSG, CREATE_REPLY, 20, &header, &len);
   may_internal_flag = 0;
   if (flag < 0) {
      perror ("mayCreate: error obtaining the create reply");
      close (open_socks->socket_id);
      tmp = open_socks;
      open_socks = open_socks->next;
      free (tmp);
      return (NULL);
   }
   if (flag == MAY_TIMED_OUT) {
      perror ("mayCreate: daemon not responding");
      close (open_socks->socket_id);
      tmp = open_socks;
      open_socks = open_socks->next;
      free (tmp);
      return (NULL);
   }
   if (cr.errortype != 0) {
      perror ("mayCreate: remote failure");
      fprintf (stderr, "%s\n", cr.err_msg);
      close (open_socks->socket_id);
      tmp = open_socks;
      open_socks = open_socks->next;
      free (tmp);
      return (NULL);
   }

#ifdef DEBUG
   printf(" maycreate recevieved acknowledge \n");
#endif

   /* allocate space for the thing to be returned. */
   if ((sa = (struct mayProcId *) malloc (SIZEOPRO)) == NULL) {
      perror ("mayCreate: error on malloc");
      return (NULL);
   }

   if (requested_connection_type == MAY_CREATE_TCP) {
      /*
       * for TCP, the child has inherited the daemons address,
       * and so, we return only that.
       */
      memcpy (sa, remhost, SIZEOPRO);
      sa->unread_bytes_remaining = 0;
      sa->partial_msg = NULL;
   }

   else if (requested_connection_type == MAY_CREATE_UDP) {
      /* 
       * if this is a UDP create, we need to return the remote (childs)
       * address to the user (parent). This address was mailed to us 
       * in the create message.  Extract it, and return.
       */
      memcpy (sa, &cr.sproc, SIZEOPRO);
      /* note, however, that we will use the *LOCAL* file descriptor
       * when mailing */
      sa->socket_id = The_udp_sock->socket_id;

      /* remove the TCP connection to the daemon */
      close (open_socks->socket_id);
      tmp = open_socks;
      open_socks = open_socks->next;
      free (tmp);

   }

   else {  
      perror ("mayCreate: unsupported connection type");
      return (NULL);
   }

   /* do not give user access to internal data areas (hostname) */
   if ((sa->hostname =
      (char *) malloc ( strlen(remhost->hostname)+1 )) == NULL) {
      perror ("mayGetTcpConnection: error on malloc");
      return (NULL);
   }
   strcpy (sa->hostname, remhost->hostname);

   return (sa);

}

/* ============================================================= */
/*
 * mayGetHosts
 * Returns address of maydaemon running on indicated host.
 *
 * This routine gets the address of the indicated service on the 
 * indicated host.  It returns a pointer to a structure containing the
 * hostname and the service address 
 *
 * Internally, it then keeps around a link list of these so
 * that it doesn't have to go through that effort again.
 *
 * It may be possible to gain a minor performance improvement by
 * converting the host list to a hash table.
 */

#ifdef ANSI_C
struct mayProcId *mayGetHosts (char *requested_hostname, 
                                 char *service)
#else
struct mayProcId *mayGetHosts (requested_hostname, service)
char *requested_hostname;
char *service;
#endif
{

   struct mayProcId *hl;
   struct hostent *he;
   struct servent *se;
   char host_name [512];   /* host name */
   int found;

   /* look for the requested host in our locally maintained host list */
   for (hl = may_hosts; hl != NULL; hl = hl->next) {
      if (!strcmp (requested_hostname, hl->hostname)) {
#ifdef DEBUG
         printf ("mayGetHosts found in its list that host %s", hl -> hostname );
         printf ( " has address %d 0x%x %d \n", hl->address.sin_family,
                            hl->address.sin_addr.s_addr, hl->address.sin_port);
#endif
         /* found it ! */
         return (hl);
      }
   }

   /* if we don't have any info about the requested host, go get that info */

   /* if the hostname was not specified, assume its the local host */
   if(requested_hostname == NULL) {

      /* get the name of the local host */
      if (0 != gethostname (host_name,  sizeof (host_name))) {
         perror ("mayGetHosts error: gethostname");
      }
   } 
   else {
      strcpy (host_name, requested_hostname);
   }

   /* get host addresses */
   he = (struct hostent *) gethostbyname (host_name);

#ifdef GORP
   if ( he == NULL) {
      /* Hmmm .. maybe its a dotted internet address ... */
      he = gethostbyaddr (host_name, strlen(host_name), AF_INET);
   }
#endif

   /* Hmmm ... still null ... */
   if (he == NULL) {
      perror ("mayGetHosts error: gethostbyname");
      return (NULL);
   }

   /* get well known service port number  */
   if((se = (struct servent *) getservbyname(service,NULL)) == NULL) {
      perror ("mayGetHosts error: getservbyname");
      return (NULL);
   }

   /* malloc space for the structure.   */
   if((hl = (struct mayProcId *) malloc (SIZEOPRO)) == NULL) {
      perror ("mayGetHosts error: malloc");
      return (NULL);
   }

   /* if you don't zero it out, you'll have socket problems !!! */
   memset (hl, 0, SIZEOPRO);
   if ((hl->hostname = (char *) malloc (strlen (host_name) + 1)) == NULL) {
      perror ("mayGetHosts error: malloc");
      return (NULL);
   }

   /* fill in the name field */
   strcpy (hl->hostname, host_name);

   /* fill in address structure     */
   hl->address.sin_family = he->h_addrtype;
   hl->address.sin_port = se->s_port;
   memcpy (&hl->address.sin_addr.s_addr, he->h_addr, he->h_length);

   /* chain in to known hosts list. */
   hl->next = may_hosts;
   may_hosts = hl;

#ifdef DEBUG
   printf ("mayGetHosts thinks that host %s", hl -> hostname );
   printf ( " has address %d %x %d \n", hl->address.sin_family,
   hl->address.sin_addr.s_addr, hl->address.sin_port);
#endif

   /* return pointer to the requested host */
   return (hl);
}


/* ============================================================= */
/* 
 * mayTerminate 
 * Closes all open sockets and terminates the may system
 *
 * --- uhhh we don't really close the file descriptors do we ???
 */

void mayTerminate ()
{
   struct mayProcId *tmp;
   struct may_msg_list *tmp_msg;

   /* close all sockets, and free associated structures */
   while (open_socks) {
      tmp = open_socks;
      open_socks =  open_socks->next;
      close (tmp -> socket_id);
      free (tmp->hostname);
      free (tmp);
   }

   /* free all memory associated with cached hosts */
   while (may_hosts) {
      tmp = may_hosts;
      may_hosts =  may_hosts->next;
      free (tmp->hostname);
      free (tmp);
   }

   /* discard all messages */
   while (may_msg_head_ptr) {
      tmp_msg = may_msg_head_ptr;
      may_msg_head_ptr = may_msg_head_ptr -> next;
      free (tmp_msg -> msg);
      free (tmp_msg); 
   }
      
}

/*----------------------end of file --------------------------*/
