/* =================================================================
 *
 * Copyright (C) 1989, 1990, 1993 Linas Vepstas
 *
 *                   All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its 
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies 
 * and that both the copyright notice and this permission notice 
 * appear in supporting documentation.
 *
 * Permission to redistribute this software and its documentation
 * is hereby granted, provided that such redistribution is not 
 * performed for monetary profit or gain.
 *
 * Redistribution for profit without express written consent is
 * prohibited.  Contact the author if you wish to use this software 
 * and/or its documentation in a commercial product.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND.  
 * THE AUTHOR DISCLAIMS ALL EXPRESS AND IMPLIED WARRANTIES, INCLUDING 
 * BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE HELD LIABLE FOR ANY SPECIAL, DIRECT, 
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING 
 * IN LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Bug reports or fixes may be sent to the author, who may or may not 
 * act upon them as he desires.
 * 
 * The author's last known address is:
 * 
 *                       Linas Vepstas
 *                     1518 Enfield Road
 *                  Austin, Texas 78703-3424
 *                      1-(512)-499-8246
 *                    linas@austin.ibm.com
 *
 * =================================================================
 */

#include <stdio.h>
#include <string.h>

#include "may.h"
#include "farm.h"

struct mayFarm *current_farm;

/* ============================================================== */
/* mayCreateFarm
 * Start multiple processes on remote processors
 *
 * This program does multiple mayCreates.  The resulting process
 * addresses are placed into a ring, hung off of a global pointer.
 * 
 * The point of this program is to provide a user convenience for
 * starting up a large number of processes that will appear to be
 * identical in many repsects.  This is ideal for a processor farm 
 * type situation.
 *
 * A linked list of the resultant creates are hung off of the global
 * variable  current_farm.
 */

#ifdef ANSI_C
void mayCreateFarm (int num_machines,	 /* number of available processors */
                    char *machine_name[],/* array containing machine names */
                    char *file_name[],	 /* array containing file paths */
                    int connection_type) /* MAY_CREATE_UDP or MAY_CREATE_TCP */
#else
void mayCreateFarm (num_machines, 
                    machine_name, file_name, connection_type)
		int num_machines;	/* number of available processors */
		char *machine_name[];	/* array containing machine names */
		char *file_name[];	/* array containing file paths */
		int connection_type;	/* MAY_CREATE_UDP or MAY_CREATE_TCP */
#endif
{
   struct mayChild *child, *tmp, *first;
   int i;

   current_farm = (struct mayFarm *) malloc (sizeof (struct mayFarm));
   current_farm -> num_workers = num_machines;

   /* num_active_workers will count how many processes actually 
    * started up.  */
   current_farm -> num_active_workers = 0;
   for (i=0; i<num_machines; i++) { 
      /* keep a record of who/what was started */
      child = (struct mayChild *) malloc (sizeof (struct mayChild));
      /* link up the first entry in the ring */
      if (i == 0) {
         first = child;
         first -> next = first;
      }
      child -> program_name = (char *) malloc (strlen (file_name[i])+1);
      strcpy (child -> program_name, file_name[i]);


      /* OK, now try to fire it up */
      child -> rpaddr = mayCreate (machine_name[i], file_name[i],
           connection_type);
      if (child -> rpaddr == NULL ) {
          fprintf (stderr, " Machine %s failed to respond \n", machine_name [i]);
          child -> alive = FALSE;
       } else {
          child -> alive = TRUE;
          current_farm -> num_active_workers++;
       }

       /* Now, link into place */
       tmp = first -> next;
       first -> next = child;
       child -> next = tmp;
   }

   current_farm -> workers = first;

}

/* =================================================================== */

#define LINE_LENGTH 255

#define CHECK_FOR_NULL(next_tok)			\
{							\
        if (next_tok == NULL) continue;			\
        if (strncmp (next_tok, "#", 1) == 0) continue;	\
        if (strncmp (next_tok, "!", 1) == 0) continue;	\
}

/* =================================================================== */
/* get_machines_and_file_paths 
 * return ASCII strings containing machine and executable paths
 *
 * This utility opens the file "machines" in the current directory, 
 * reads the machine names and executable file paths contained therein, and
 * returns them.  Note that the arguments are addresses of pointers to
 * lists of pointers -- its just like argv, except that you have to pass
 * the ADDRESS of argv to this routine when invoking it.  See examples
 * for usage.
 */


#ifdef ANSI_C
int get_machines_and_file_paths (char **machine_name[],/* array of machine names */
                                 char **file_name[]) /* array of file paths */
#else
int get_machines_and_file_paths (machine_name, file_name)
char 	**machine_name [];	/* array containing machine names */
char	**file_name [];	/* array containing file paths */
#endif
{
    FILE *fyle;
    char input_record [LINE_LENGTH];
    char *next_tok;
    char **ma, **fi;
    int len;
    int n=0;

    /* open file containing machine and file names */
    if ((fyle = fopen ("./machines", "r")) == NULL) {
        perror ("file open failure: can't find: machines");
        exit (EXIT_FAILURE);
    }
   
   /* alloc arrays of pointers to names */
   ma = (char **) malloc (100*sizeof(char *));
   fi = (char **) malloc (100*sizeof(char *));
   *machine_name = ma;
   *file_name = fi;

    /* loop forever, until we break out of loop */
    while (TRUE) {

        /* if end of file or read error, return */
        if (!fgets (input_record, LINE_LENGTH, fyle)) {
           fclose (fyle);
           ma [n] = NULL;
           fi [n] = NULL;
           return (n);
        }

        /* if input line longer than buffer, ignore it. */
        if (strlen (input_record) == LINE_LENGTH - 1) {
           perror ("file read failure: line too long: machines");
           continue;
        }

        /* ignore comment cards */
        if (strncmp (input_record, "#", 1) == 0) continue;
        if (strncmp (input_record, "!", 1) == 0) continue;

        /* get machine name -- if no tokens found, or blank line try again */
        next_tok = strtok (input_record, " ,;\t\n\r");
        if (next_tok == NULL) continue;
        if (strncmp (next_tok, "#", 1) == 0) continue;
        if (strncmp (next_tok, "!", 1) == 0) continue;
        len = strlen (next_tok) + 1;
        ma[n] = (char *) malloc (len * sizeof (char));
        strcpy (ma[n], next_tok);

        /* get the file name */
        next_tok = strtok (NULL, " ,;#!\t\n");
        if (next_tok == NULL) continue;
        if (strncmp (next_tok, "#", 1) == 0) continue;
        if (strncmp (next_tok, "!", 1) == 0) continue;
        len = strlen (next_tok) + 1;
        fi[n] = (char *) malloc (len * sizeof (char));
        strcpy (fi[n], next_tok);

        /* go do the next pair */
        n++;

        /* if the array size is exceeded, malloc more memory */
        if (n%100 == 0) {
            ma = (char **) realloc (ma, n*sizeof (char *));
            fi = (char **) realloc (fi, n*sizeof (char *));
            *machine_name = ma;
            *file_name = fi;
        }

    }

    /*  Below will accomplish same as above, but would be 'hardwired'
     * into code.
    strcpy (machine_name [0], "taupe"); strcpy (file_name [0], "/tmp/dslave");
    strcpy (machine_name [1], "khaki"); strcpy (file_name [1], "/tmp/dslave");
    strcpy (machine_name [2], "okra"); strcpy (file_name [2], "/tmp/dslave");
    strcpy (machine_name [3], "");
     */
}

/* =========================== END OF FILE ============================== */
