
/* =================================================================
 *
 * Copyright (C) 1989, 1990, 1993 Linas Vepstas
 *
 *                   All Rights Reserved
 *
 * Permission to use, copy, and modify this software and its 
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies 
 * and that both the copyright notice and this permission notice 
 * appear in supporting documentation.
 *
 * Permission to redistribute this software and its documentation
 * is hereby granted, provided that such redistribution is not 
 * performed for monetary profit or gain.
 *
 * Redistribution for profit without express written consent is
 * prohibited.  Contact the author if you wish to use this software 
 * and/or its documentation in a commercial product.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND.  
 * THE AUTHOR DISCLAIMS ALL EXPRESS AND IMPLIED WARRANTIES, INCLUDING 
 * BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF MERCHANTABILITY AND 
 * FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL THE AUTHOR BE HELD LIABLE FOR ANY SPECIAL, DIRECT, 
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING 
 * IN LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Bug reports or fixes may be sent to the author, who may or may not 
 * act upon them as he desires.
 * 
 * The author's last known address is:
 * 
 *                       Linas Vepstas
 *                     1518 Enfield Road
 *                  Austin, Texas 78703-3424
 *                      1-(512)-499-8246
 *                    linas@austin.ibm.com
 *
 * =================================================================
 */

#define TRACE

#ifdef TRACE
#include <math.h>
#endif

#include "may.h"
#include "farm.h"

/* ============================================================== */
/* mayProcFarm
 * Implements a processor farm for distributing work in parallel
 *
 * This routine implements algorithms for distributing a large number
 * of pieces of work (pieces, tasks, jobs) across a number of remote
 * processors.  A number of tricks have been squeezed into this
 * routine-- tricks for load balancing, for fault-tolerant
 * distribution of the work, a fast-feed mode for exceptionally fast
 * processors (or exceptionally small jobs), and all this optimized
 * so that this routine executes with minimal overhead.
 *
 * The underlying assumptions are these: (1) the pieces can be 
 * executed in arbitrary order.  The load balancing mechanisms 
 * assume (2) the size of all job pieces are roughly equivalent 
 * (to within an order of magnitude, or so.)
 *
 * Note that some extra code is interpersed to monitor statistics.
 *
 * The high water mark is implemented on a per machine basis.
 * What is the high water mark?  It is the maximum amount of work
 * that should ever be queued up at that machine.  Why should
 * each machine have a separate high-water mark?  Because we
 * want to use it for load balancing.  Each machine should have
 * the same amount of (capability-adjusted) work queued up, so 
 * that they all would complete in the same amount of time.
 * In other words, fast machines should have more work queued up
 * than slow machines.  
 *
 * Note that the high-water mark is dynamically adjusted. I.E. we
 * don't know 'a priori' how fast any given machine is.  What 
 * we do is keep tabs on how much work it has done, relative to 
 * the other machines.  We adjust high water mark accordingly.
 *
 * The low water mark is also done on a per machine basis so that
 * exteremly efficient machines can be kept busy.  If a fast
 * machine is able to drain from low-water to zero faster than
 * we can get back to it and fill it up again, it will sit idle.
 * This can occur due to either the structure of this algorithm,
 * or due to network propagation delays, etc.  Keeping the low water
 * mark up will avoid idle time on the remote machine.
 *
 * Note that the low-water mark is also dynamically adjusted. It is
 * bumped up by a different method than the high-water mark. It
 * is bumped up every time a queue is (accidentaly) drained to 
 * zero, indicating that we didn't service that machine soon
 * enough.
 *
 */

    /* A variable timeout mechanism implemented.  This mechanism adjust
for fastest machine (actually to wieghted avg) */


#ifdef ANSI_C
int mayProcFarm (Void * work,	/* pointer to array of pieces to be done */
		int num_pieces,	/* number of pieces to be done */
		int piecesize,	/* size (in bytes) of each piece */
		int stride,	/* spacing (in bytes) 'tween pieces */
		int ((*callback)())) /* called back when piece is done */
#else
int mayProcFarm ( work, 
                num_pieces,
		piecesize, 
		stride, 
		callback) 

Void * work;		/* pointer to array of pieces to be done */
int num_pieces;		/* number of pieces to be done */
int piecesize;		/* size (in bytes) of each piece */
int stride;		/* spacing (in bytes) between subsequent pieces */
int ((*callback)());	/* routine called back every time a piece is done */
#endif

{
    unsigned int	iarr;

    int i;

    int num_machines;	/* the number of available processors */

    struct mayMsghead	header;		/* header of returned message */
    int		messlen;		/* length of returned message */
    static float data [2048];		/* my data array */
    int		retval;			/* returned value */

    long	low_water_min;		/* min value that low water can take */
    long	low_water_max;		/* max value that low water can take */
    long	high_water_max;		/* max value that high water can take */
    unsigned int   batch;		/* job batching counter */
    unsigned int   batch_min;    /* minimum num of msgs sent to each mach */
    unsigned int   batch_block;  /* num of messages sent to each machine */
    unsigned int   num_free_machines;	/* number of free processors */
    unsigned int   curr_machine;	/* load balancing pointer */
    double         panic_timeout;        /* seconds to wait before panicking */
    unsigned int   timeout_period;	/* how long to wait before timing out*/
    unsigned int   max_backlog_factor;	/* max allowed accumulated backlog */
    unsigned int   wait_for_work;	/* T or F, wait before timing out*/
    unsigned int   do_receive_loop;	/* receive fault flag */
    unsigned int   not_done;		/* finished flag */
    unsigned int   num_finished;		/* number of completed tasks */
    unsigned int   num_since_last_transmit; /* num received since last timeout*/
    unsigned int   num_sessions;	 /* number of timeouts to completeion */
    int   instant_refill;
    unsigned int   icnt;
    unsigned int   i_last_looked_at;

#ifdef TRACE
    /* variables used only for performance monitoring and trace */
    unsigned int   num_duplicated; /* num of tasks mistakenly duplicated */
    unsigned int   percent_dups;	/* percent of duplicated tasks */
    unsigned int   num_sent;		/* number of sent tasks */
    int   dribble_start;	/* num of timeouts before dribble */
    unsigned int   num_done_pre_dribble; /* num jobs completed before dribble*/
    unsigned int   num_dup_pre_dribble; /* num duplicates before dribble*/
    long	elapsed_time;		/* elapsed time */
    int		seconds, minutes, hours; /* elapsed time */
    unsigned int   unique_done;		/* amount of unique work done */
    int		 num_empties;    /* number of times a zero-length queue encountered */
    float 	avg;		 /* temp variable */
#endif

    struct WorkStatus {	/* workpiece status array */
	int done;	/* if false, then piece has been done */
    } *workstatus;
    
    struct MachineStatus { 	/* machine status array */
        long	queue_length;	/* msg queue length at each machine */
        short	is_free;	/* TRUE if queue low and to be filled */
        long	num_done;	/* pieces done by this machine */
	int	high_water_mark;/* maximum queue length */
	int	low_water_mark; /* minimum queue length */
#ifdef TRACE
        int   sum_queue_len;		/* sum of queue lengths */
        long  sum_square_len;		/* sum of squares of queue lengths */
#endif
    } *machstat;

    struct mayProcId **rem_proc; /* array of available, initialized procs */
    struct mayChild *tmp;

    /* get stuff out of global data area */
    num_machines = current_farm -> num_active_workers;

    if (num_machines <= 0) {
       fprintf(stderr, "mayProcFarm: no processors specified \n");
       return (MAY_NO_OPEN_SOCKS);
    }

    rem_proc = (struct mayProcId **) 
         malloc (num_machines * sizeof (struct mayProcId *));
    if (rem_proc == NULL) {
       perror(" mayProcFarm: unable to malloc");
       return (MAY_NO_MEM);
    }

    tmp = current_farm -> workers;
    for (i=0; i<num_machines;) {
        if (tmp->alive) {
            rem_proc[i] = tmp -> rpaddr;
            i++;
        }
        tmp = tmp -> next;
    }


    /* ------------------------------------------------
     * Set some very important and relevent parameters.
     * These affect system response and thouroughput.
     * See below for explanation.
     */
    panic_timeout = 1.01;	  /* seconds to wait before timing out */
    timeout_period = (unsigned int) panic_timeout;

    high_water_max = (long) 
             (0.50 * ((float) num_pieces) / ((float) num_machines));
    if (high_water_max <10) high_water_max = 10;

    low_water_min = 3;
    low_water_max = (3 * high_water_max) / 4;
    max_backlog_factor = 2;
    batch_min = 5;
    /* ------------------------------------------------ */

   /* set up and initialize data areas */
   workstatus = (struct WorkStatus *) 
      malloc (num_pieces * sizeof(struct WorkStatus));
   if (NULL == workstatus) {
      perror(" mayProcFarm: unable to malloc");
      return (MAY_NO_MEM);
   }

   for (i=0; i<num_pieces; i++) {
      workstatus[i].done = FALSE;
   }

   /* intialize per-machine stuff */
   machstat = (struct MachineStatus *) 
      malloc (num_machines * sizeof(struct MachineStatus));
   if (NULL == machstat) {
      perror(" mayProcFarm: unable to malloc");
      return (MAY_NO_MEM);
   }

   for (i=0; i<num_machines; i++) {
       machstat[i].queue_length = 0;
       machstat[i].is_free = TRUE;
       machstat[i].num_done = 0;
       machstat[i].high_water_mark = high_water_max;
       machstat[i].low_water_mark = low_water_min;
#ifdef TRACE
       machstat[i].sum_queue_len = 0;
       machstat[i].sum_square_len = 0;
#endif
    }
#ifdef TRACE
    /* start measuring elapsed time after the window has been placed */
    elapsed_time = time (NULL);
    printf (" mayProcFarm: completed initializing \n");
#endif

    num_free_machines = num_machines;
    curr_machine = 0;
    num_finished = 0;
    num_since_last_transmit = 0;
    num_sessions = 0;
    instant_refill = 0;

#ifdef TRACE
    printf ("\n\n");

    /* start the data transmission and collection loop */
    num_duplicated = 0;
    unique_done= 0;
    dribble_start = -1;
    num_empties = 0;
    num_sent = 0;
#endif

    /* start the data transmission and collection loop */
    wait_for_work = timeout_period;
    not_done = TRUE;
    icnt = 0;
    while (TRUE) {
       num_since_last_transmit = 0;
       /*
        * Send out a big glob of data, in round-robin fashion, to all
        * participating machines.   
        *
        * Note that for a UDP-based farm, a lot of messages will get
        * dumped.  The problem here is that UDP does not guarentee 
        * delivery of messages, and the maySend routine is very capable 
        * of dumping so many messages that the net device driver will 
        * croak, resulting in a bunch of lost messages. Some smaller 
        * number of messages will get lost on congested gateways, etc.
        * 
        * This dumping is not a problem for TCP, since TCP gaurentees
        * delivery. TCP also has a bunch of high tech built into it to 
        * avoid network congestion, so a TCP-based farm is likely to 
        * be freindlier to a loaded network/machine. Its up to you to 
        * figure out if a UDP or TCP based farm results in a faster 
        * overall execution time.
        *
        * The size of the glob of data  sent out is given by the
        * "batch_block" variable.
        */

       /* count the number of send-recv sessions */
       num_sessions ++;

       /* determine how much to send out, based on the number we are 
        * losing on the network */
       batch_block = batch_min + num_finished/(num_sessions*num_machines);

       i_last_looked_at = icnt;
       for (i=0; i < batch_block*num_free_machines; ) {

          /* skip this machine if its queue is full */
          while (machstat[curr_machine].is_free == FALSE) {
             curr_machine ++;
             curr_machine %= num_machines;
          }

          if (workstatus[icnt].done == FALSE) {
             i++;
             maySend (rem_proc [curr_machine], 37+curr_machine, (short) icnt, 
                   ((Void *) (((char *) work) + icnt*stride)), piecesize );
             curr_machine ++; 
             curr_machine %= num_machines;
#ifdef TRACE 
             num_sent ++;
#endif
          }

          icnt ++;
          icnt %= num_pieces;
          if ( icnt == i_last_looked_at ) {
              /* if icnt equals last-looked-at, we've come around once.
               * if i is zero, means we didn't find anything that wasn't
               * done. So we're outa here.
               */
              if ( i == 0 ) not_done = FALSE;
#ifdef TRACE
              /* keep a fresh set of statistics for the tail end of work */
              if (dribble_start < 0) {
                  dribble_start = num_sessions;
                  num_done_pre_dribble = num_finished;
                  num_dup_pre_dribble = num_duplicated;
              }
#endif
              break;
          }
       }

       if (!not_done) break;

       /* adjust the queue lengths downwards as we approach the end.
        * (Note this is still a kludge).
        */
       if (num_finished > ((int) (0.8 * (float) num_pieces))) {
          for (i=0; i<num_machines; i++) {
             machstat[i].high_water_mark /=2;
             if (machstat[i].high_water_mark < 3)
                machstat[i].high_water_mark = 3;
          }
       }



       /* 
        * Ok, now receive the completed work.
        * The plan here is that we read until at least one machine
        * has gotten to its low water mark, at which point we go
        * out to fill it back up.  
        * 
        * If this program is implemented on top of a reliable
        * communications medium (e.g. TCP/IP), this would be enough.
        * However, we are using unreliable (UDP/IP) datagrams.
        * The net effect is that we don't have a reliable model of the 
        * queue lengths at the remote machines.  In particular:
        *     
        * --> packets containing work done (and also the queue
        *     length indicator) might be lost. In particular,
        *     a fast processor may have drained its queue, yet
        *     have had every message lost. Thus, our local model
        *     thinks this guys queue is above high water when in fact
        *     its at zero.
        *     
        * --> extermely slow processors might be issueing stale reports
        *     (i.e. reporting short queues, although the actual queue
        *     is longer because it has been recently filled).
        *
        * Thus we need a "sanity check" -- an alternate mechanism for
        * judging processor activity, and the need for refilling
        * queues.  The alternate mechanism implemented below is a 
        * variable, dynamically adjusted, timeout mechanism.
        * Basically, we "monitor" the rate of arrival of messages.
        * As long as messages keep arriving, we leave well-enough 
        * alone.  However, if a period of "timeout" elapses with
        * no recieved messages, we assume something has gone wrong
        * and go off and feed the machines. (We feed all machines,
        * irrespective of the current (but possibly inaccurate) 
        * settings of the water-marks/queue lengths).
        * 
        * The $6,000 question is, what should the timeout period be ?
        * The answer seems to be: we need TWO differnet timeout periods.
        * The first is called a "short timeout", which we use to fill
        * the processor queues.  That is, we timeout relatively
        * frequently in order to fill up queues. (Actually, this is a
        * bad idea, since it may lead to extremely long queues on slow
        * processors).
        *
        * The other timeout is a "panic timeout", which, if nothing
        * (much) happens during it, we panic, and assume a malfunction.
        * Panicking consists of sending ALL machines more fresh work
        * to do (and then resuming to normal operation).
        *
        * In order to have a robust algorithm, we need a mechanism that
        * can dynamically adjust for large or small work pieces.  If
        * we timeout infrequently, we risk having the remote machines
        * sitting idle, and thus a large elapsed time for execution.
        * If we time out excessively frequently, we risk having large
        * percentages of duplicated jobs (because jobs in some
        * queue will be mistaken for unassigned jobs, and be farmed out
        * to someone else).
        *
        * Note also, that with short timeouts, we risk overfilling 
        * queues. For example, assume all processors are equal
        * in power, and that the time to execute one job is very long
        * compared to the timeout period.  Since we can not know the
        * length of the queue at some processor until at least one job
        * is returned, timing out often will result in extremely large
        * numbers of jobs being sent to the processors. (This point 
        * applies only to UDP (unreliable) transmission mediums).
        *
        * If, on the other hand, the communications medium is reliable,
        * we can predict the length of the queue at the remote processor
        * very easily (it is just the number of messages that we sent).
        * In such a case, we may as well use an infinite timeout until
        * at least one queue has drained below the low water mark,
        * at which point we go to zero timeout, so that we can break
        * out and go back up to fill the queue.
        *
        * Therefore, we re-evaluate the timeout period, basing it on
        * the number of messages received since the last timeout.
        * Note, however, we want to risk on the side of timing out too
        * often, rather than not often enough.  Therfore, 1.55 is
        * smaller than 2.0.  Note also, that 1.55 is slightly larger than
        * the square root of 2.0.
        */

       header.msgptr = (char *) &data[0];

       instant_refill --;

       do_receive_loop = TRUE;
       while (do_receive_loop == TRUE) {
          retval = mayReceive (NULL, NULL, NULL, 
                                   wait_for_work, &header, &messlen);
          if (retval == MAY_TIMED_OUT) {
             do_receive_loop = FALSE;

             /* update our geusstimate for the timeout period only
              * if we timed out from natural causes (and not from
              * low water marks).  (Note: the wait_for_work variable
              * in the if statement below is also in the mayReceive
              * call above.  Normally, it is a positive number, except
              * when a machine is below the low water mark, in which 
              * case it is zero.)
              */
             if (wait_for_work != 0) {
#ifdef TRACE
                printf ("Timed out on read, after %d secs \n", wait_for_work);
                printf(" adjust timeout, old= %f", panic_timeout);
#endif
                /* 8 and 24 were choosen because they put us in a
                 * reasonable range for Poisson statistics --
                 * In other words, if, on average we receive at least 8
                 * messages, chances are tiny that we'll time out by accident.
                 * On the other hand, if we receive more than 24, we're 
                 * probably waiting to long to long, and are not
                 * dumping new work out on the net often enough.  
                 */
                if (num_since_last_transmit < 8 ) panic_timeout *= 1.55;
                if (num_since_last_transmit > 24 ) panic_timeout /= 2.0;
                /* make sure its never less than a second */
                if (panic_timeout < 1.01) panic_timeout = 1.01;
                if (num_finished == 0) {
                   panic_timeout *= 3.0/1.55;
                   if (panic_timeout > 3600.0) panic_timeout = 3600.0;
                }
#ifdef TRACE
                printf(" new = %f \n", panic_timeout);
#endif
                timeout_period = (unsigned int) panic_timeout;

                /* Since we timed out, we can safely assume that some
                 * and/or all machines are dead.  Lets make sure we 
                 * transmit to all.  Mark them as all being free.
                 */
                for (i=0; i<num_machines; i++) {
                   machstat[i].is_free = TRUE;
                } 
                num_free_machines = num_machines;
             }

             /* Reset the wait period for the next time through.
              * But do this only if we are not in a force-feed it
              * mode.  Reminder: we enter force-feed mode whenever a 
              * machine drops to zero queue length.  Note we are in
              * force-feed mode whenever instant refill is non-zero.
              */
             if (instant_refill <= 0) {
                wait_for_work = timeout_period;
                instant_refill = 0;
             }
          }  /* endif TIMED_OUT */
          else {

             num_finished++;
             num_since_last_transmit++;
             /* 
              * Mark this data segment as having been processed 
              */
             iarr = (unsigned int) header.subtype;
#ifdef TRACE
             /* If its already been processed, count it as a duplicate. */
             if (workstatus[iarr].done == TRUE) {
                num_duplicated ++;
             } else {
                unique_done ++;
             }
#endif
             /* mark this piece of work as having been completed */
             /* if the user supplied a callback, give it the data */
             /* but do so only if the piece had never been done before */
             if (workstatus[iarr].done == FALSE) {
                if (callback != NULL) (*callback) (data);
             }

             workstatus[iarr].done = TRUE;
             i = header.msgtype - 37;

             /* machine i has done this work */
             machstat[i].num_done ++;

             /*
              * disable machines which have plenty of work queued up,
              * enable machines that almost done with thier stuff.
              *
              * Note that if at least one machine is below its low 
              * water mark, we will want to be filling it up again
              * sometime soon. We do this by setting the timeout
              * to zero, so that we finish processing any/all messages
              * that have been received, and then go refill.
              *
              * Note that there is also a panic mode: if some queue
              * reports zero length, we refill IMMEDIATELY (instantly),
              * we don't bother to finish processing any work that is 
              * pending (because that may take too long).
              *
              * On the other hand, it may be possible that the reason
              * these remote machines are going idle is because the 
              * master process (i.e. the callback) is taking too long.  
              * We can detect this by noting that there is a backlog 
              * of unfinished work in our read queue.  We shouldn't
              * let this backlog get too high, because the panic-timeout
              * mechanism could result in a flood of duplicated work 
              * requests, slowing down overall execution time.
              *
              * Note that if all machines are above the high water mark,
              * we will kick back and not bother to
              * go back to the send loop until someone gets below
              * the low water mark.
              *
              */
             if (header.queue_len < machstat[i].low_water_mark) {
                 /* if simply low, keep reading until no-wait */
                 wait_for_work = 0; 
                 if (machstat[i].is_free == FALSE) {
                     num_free_machines ++;
                     machstat[i].is_free = TRUE;
                 }

                 /* if queue down to zero, feed it NOW!!! */
                 if ((header.queue_len == 0) && 
                        (retval < max_backlog_factor * num_machines)) {
                    do_receive_loop = FALSE;
                    instant_refill = 5;
#ifdef TRACE
                    /* Count the number of times some machine bottomed out */
                    num_empties ++;
                    printf (" mayProcFarm: zero queue length on machine %d \n", i);
#endif
                    /* 
                     * but we will also raise the low water, so that it
                     * don't happen again 
                     */
                    machstat[i].low_water_mark *=2;
                    if (machstat[i].low_water_mark > low_water_max) {
                        machstat[i].low_water_mark = low_water_max;
                    }
                    if (machstat[i].low_water_mark > 
                                machstat[i].high_water_mark) {
                        machstat[i].high_water_mark *=2;
                        if (machstat[i].high_water_mark > high_water_max) {
                            machstat[i].high_water_mark = high_water_max;
                        }
                    }
printf (" mach %d low= %d hi = %d \n", i, machstat[i].low_water_mark,
 machstat[i].high_water_mark);
                }  /* endif (instant refill check) */   
             }

             if (header.queue_len > machstat[i].high_water_mark) {
                 if (machstat[i].is_free == TRUE) {
                     num_free_machines --;
                     machstat[i].is_free = FALSE;
                 }
                 if (num_free_machines == 0) wait_for_work = timeout_period;
             }

#ifdef TRACE
             /* 
              * Count the amount of work done,
              * Keep stat on which machine is doing what work.
              */
             machstat[i].sum_queue_len += header.queue_len;
             machstat[i].sum_square_len += header.queue_len * header.queue_len;
#endif

          }
       }

#ifdef TRACE
       printf ("mayProcFarm: session %d, ", num_sessions);
       printf ("batch= %d ", batch_block );
       printf ("done= %d", num_since_last_transmit );
       printf (" total= %d unique= %d \n", num_finished, unique_done);
#endif

    }
 
#ifdef TRACE
    printf ("\n\n\n");

    elapsed_time = time(NULL) - elapsed_time;
    seconds = elapsed_time %60;
    elapsed_time /= 60;
    minutes = elapsed_time %60;
    elapsed_time /= 60;
    hours = elapsed_time;
    printf (" elapsed time = %d h %d m %d s \n", hours, minutes, seconds);

    printf (" total number of pieces needed = %d \n", num_pieces);
    printf (" total number of pieces done = %d \n", num_finished);

    percent_dups = (100 * num_duplicated) / (num_finished - num_duplicated);
    printf (" percent duplicated jobs %d \n", percent_dups);

    printf (" number of messages sent per machine per batch = %d \n",
                               batch_block);

    avg = (float) num_sent / (float) num_sessions;
    printf (" average number of messages sent per batch = %f \n", avg);

    avg = (float) num_finished / (float) num_sessions;
    printf (" average number of messages received per batch = %f \n", avg);

    printf (" total number of batches sent = %d \n", num_sessions);

    percent_dups = (100 * (num_sent - num_finished)) / num_finished;
    printf (" percent of msgs lost on network = %d \n", percent_dups);

    printf ("\n");
    printf (" number of batches sent before petering out at end = %d \n",
                               dribble_start);
    if (dribble_start > 0) {
        printf (" avg num of msgs recvd per batch (steady state mode) = %f \n",
          (float) num_done_pre_dribble / (((float)dribble_start) +0.001) );

        percent_dups = (100 * num_dup_pre_dribble) / 
                (num_done_pre_dribble - num_dup_pre_dribble+1);
        printf(" percent duplicated jobs before dribble %d \n", percent_dups);
    }

    printf (" number of batches dribbled back at the end = %d \n", 
                               num_sessions - dribble_start);
    printf (" avg num of msgs recvd per batch (final dribble) = %d \n",
      (num_finished - num_done_pre_dribble) / (1+ num_sessions - dribble_start));

    percent_dups = (100 * num_duplicated - num_dup_pre_dribble) / 
                (1+num_finished - num_done_pre_dribble - 
                     (num_duplicated - num_dup_pre_dribble));
    printf(" percent duplicated jobs in dribble %d \n", percent_dups);

    printf ("\n");

    for (i = 0; i < num_machines; i++) {
       printf (" %d did %d pieces, or %d percent of the jobs\n", i,
          machstat[i].num_done,
          (100 * machstat[i].num_done) /num_finished );
    }

    printf ("\n");
    for (i = 0; i < num_machines; i++) {
       float avg, rms;

       avg = ((float) machstat[i].sum_queue_len) / 
                                ((float) machstat[i].num_done);
       rms = ((float) machstat[i].sum_square_len) / 
                                ((float) machstat[i].num_done);
       rms -= avg*avg;
       rms = sqrt (rms);
       printf (" %d had average queue length of %f, rms change %f \n",
         i, avg, rms);
       printf (" %d had had hi watermark at %d low water at %d \n",
         i, machstat[i].high_water_mark, machstat[i].low_water_mark);
/*
       printf (" %s had average queue length of %f, rms change %f \n",
          rem_proc[i] -> hostname, avg, rms);
*/
    }
    printf (" number of times queue totally emptied %d\n", num_empties);
#endif

   return (0);

}
/* =========================== END OF FILE ======================= */
